
/* TSAY SOUND CONVERTER (c) 02/1996 Alvin Albrecht */


#include <stdio.h>
#include <stdlib.h>

#define MAXPARTS     26                     /* Max # files created */
#define FNAMELEN     7                      /* Max filename length-1 */
#define Fc           3529500.0              /* TS2068's clock rate */

int MAXSIZE=7*5*1024;                       /* Max output filesize in bytes */
                                            /* Set for 7 cyls on JLO disk*/
typedef struct
{
   char *string;
   char rate;
   char channels;
} HEADER;

int HI(int x)
{
   return((int)(x/256));
}

int LO(int x)
{
   return(x-256*HI(x));
}

float SRATE(int brate,int channels)
{
   return(Fc/(13*brate+81*channels-29));
}

int BRATE(float rate,int channels)
{
   return((int)((Fc/rate-81*channels+29)/13+0.5));
}

void error(char *message)
{
    fprintf(stderr,"%s",message);
    exit(1);
}

int strlen(char *s)
{
   int i;

   for (i=0; *s++; i++) ;
   return(i);
}

char *strncpy(char *to,char *from,int n)   /* Library function doesn't */
{                                          /* add a NULL if max # characters */
   char *temp;                             /* to copy is exceeded. */

   for (temp=to; (*from) && (n>0); n--)
      *temp++=*from++;
   *temp='\0';

   return(to);
}

void MonoConvert(FILE *infile,HEADER *info,char *name)
{
   char fname[FNAMELEN+5],*buffer;
   int i,inch,sampleblock,bufflen,last,count,slen,sidx,nybble;
   FILE *outfile;

   if ((buffer=(char *)(malloc((MAXSIZE+10)*sizeof(char))))==NULL)
      error("\nNo memory!\n\n");

   for (i=inch=0; (i<MAXPARTS) && (inch!=EOF); i++)
   {
      sampleblock=last=0;
      bufflen=strlen(strncpy(buffer,info->string,MAXSIZE))+1;
      buffer[bufflen++]=info->rate;
      buffer[bufflen++]=info->channels;

      do
      {
         for (count=1; (last==(((inch=fgetc(infile))>>4)&0xf)) && (inch!=EOF); count++) ;

         if (!sampleblock)
         {
            slen=nybble=0;
            buffer[bufflen++]=0xc0;
            sidx=bufflen;
            bufflen+=2;
            sampleblock=1;
         }

         if (!nybble)
            buffer[bufflen]=(last<<4)&0xf0;
         else
         {
            buffer[bufflen++]|=last;
            slen++;
         }
         nybble=!nybble;
         count--;

         while (count>0)
         {
            if (count<11)
            {
               while ((count--)>0)
               {
                  if (!nybble)
                     buffer[bufflen]=(last<<4)&0xf0;
                  else
                  {
                     buffer[bufflen++]|=last;
                     slen++;
                  }

                  nybble=!nybble;
               }
            }
            else if (nybble)
            {
               count--;
               nybble=0;
               buffer[bufflen++]|=last;
               slen++;
            }
            else if (count>65540)
               count=65540;
            else
            {
               buffer[sidx]=LO(slen);
               buffer[sidx+1]=HI(slen);
               sampleblock=0;

               if (count<67)
                  buffer[bufflen++]=0x40+((count-3)&0x3f);
               else
               {
                  buffer[bufflen++]=0x80;
                  buffer[bufflen++]=LO(((count-4)&0xffff));
                  buffer[bufflen++]=HI(((count-4)&0xffff));
               }
               count=0;
            }
         }

         last=((inch>>4)&0xf);

      } while ((inch!=EOF) && (bufflen<MAXSIZE));

      if (nybble)
      {
         buffer[bufflen]|=(buffer[bufflen]>>4)&0xf;
         bufflen++; slen++;
      }
      if (sampleblock)
      {
         buffer[sidx]=LO(slen);
         buffer[sidx+1]=HI(slen);
      }

      if (name[0])
      {
         sprintf(fname,"%s%c.ay",name,i+'A');
         if ((outfile=fopen(fname,"wb"))==NULL)
         {
            fprintf(stderr,"\nCan't open %s for writing!\n\n",fname);
            free(buffer);
            exit(1);
         }
      }
      else
         outfile=stdout;

      for (count=0; count<bufflen; count++)
         fprintf(outfile,"%c",buffer[count]);
      fprintf(outfile,"%c",'\0');
      if (name[0]) fclose(outfile);
   }
   free(buffer);
}

void StereoConvert(FILE *infile,HEADER *info,char *name)
{
   int i,inch,inch1,inch2,sampleblock,bufflen,last,count,sidx,slen;
   char fname[FNAMELEN+5],*buffer;
   FILE *outfile;

   if ((buffer=(char *)(malloc((MAXSIZE+10)*sizeof(char))))==NULL)
      error("\nNo memory!\n\n");

   for (i=inch1=inch2=0; (i<MAXPARTS) && (inch1!=EOF) && (inch2!=EOF); i++)
   {
      sampleblock=inch=0;

      bufflen=strlen(strncpy(buffer,info->string,MAXSIZE))+1;
      buffer[bufflen++]=info->rate;
      buffer[bufflen++]=info->channels;

      do
      {
         last=inch;
         for (count=0; ((inch1=fgetc(infile))!=EOF) && ((inch2=fgetc(infile))!=EOF) && (last==inch); count++)
            inch=(inch1&0xf0)|((inch2>>4)&0xf);

         if (!sampleblock)
         {
            buffer[bufflen++]=0xc0;
            sidx=bufflen;
            bufflen+=2;
            sampleblock=1;
         }

         buffer[bufflen++]=last;
         slen++; count--;

         while (count>0)
         {
            if (count<6)
            {
               while ((count--)>0)
               {
                  buffer[bufflen++]=last;
                  slen++;
               }
            }
            else if (count>65538)
               count=65538;
            else
            {
               buffer[sidx]=LO(slen);
               buffer[sidx+1]=HI(slen);
               sampleblock=0;

               if (count<66)
                  buffer[bufflen++]=0x40+((count-2)&0x3f);
               else
               {
                  buffer[bufflen++]=0x80;
                  buffer[bufflen++]=LO(((count-2)&0xffff));
                  buffer[bufflen++]=HI(((count-2)&0xffff));
               }
               count=0;
            }
         }
      } while ((inch1!=EOF) && (inch2!=EOF) && (bufflen<MAXSIZE));

      if (sampleblock)
      {
         buffer[sidx]=LO(slen);
         buffer[sidx+1]=HI(slen);
      }

      if (name[0])
      {
         sprintf(fname,"%s%c.ay",name,i+'A');
         if ((outfile=fopen(fname,"wb"))==NULL)
         {
            fprintf(stderr,"\nCan't open %s for writing!\n\n",fname);
            free(buffer);
            exit(1);
         }
      }
      else
         outfile=stdout;

      for (count=0; count<bufflen; count++)
         fprintf(outfile,"%c",buffer[count]);
      fprintf(outfile,"%c",'\0');
      if (name[0]) fclose(outfile);
   }
   free(buffer);
}

char *argument(char ***pargv,int *pargc)
{
   char *temp;

   if (!(*++(*pargv)[0]))
   {
      (*pargc)--;
      (*pargv)++;
   }

   temp=(*pargv)[0];
   while (*(*pargv)[0])
      (*pargv)[0]++;
   (*pargv)[0]--;

   return(temp);
}
   
main(int argc,char **argv)
{
   int channels=1,verbose=0,temp;
   float rate=8000.0;
   HEADER info;
   FILE *in;
   char outfile[FNAMELEN+1]="",infile[FNAMELEN+5+1]="";
   char name[105+1]="TSAY ";
   char c;

   if (argc>1)
   {
      if ((*(argv+1))[0]!='-')
      {
         strncpy(infile,*++argv,FNAMELEN+5);
         argc--;
      }

      while ((--argc>0) && ((*++argv)[0]=='-'))
         while (c=*++argv[0])
            switch (c)
            {
               case 'V':  verbose=1;
                          break;
               case 'f':
               case 'c':  if (sscanf(argument(&argv,&argc),"%d",((c=='f') ? (&MAXSIZE):(&channels)) )==EOF)
                             error("\nBad Integer\n\n");
                          break;
               case 'r':  if (sscanf(argument(&argv,&argc),"%f",&rate)==EOF)
                             error("\nBad Float\n\n");
                          break;
               case 'n':  strncpy(name+5,argument(&argv,&argc),100);
                          break;
               default :  error("\nUnrecognized Option\n\n");
                          break;
            }

      if (argc>0)
      {
         strncpy(outfile,*(argv++),FNAMELEN);
         argc--;
      }

      if (argc>0)
         error("\nExtra Arguments\n\n");

      if ((channels<1) || (channels>2))
         error("\nSample Player only supports mono and stereo sound\n\n");

      temp=BRATE(rate,channels);
      if ((temp<1) || (temp>256))
         error("\nSample rate out of range\n\n");

      if (MAXSIZE<10) MAXSIZE=10;

      info.string=name;
      info.rate=temp&0xff;
      info.channels=channels;

      if (infile[0])
      {
         if ((in=fopen(infile,"rb"))==NULL)
         {
            fprintf(stderr,"\nCan't open %s for reading!\n\n",infile);
            exit(1);
         }
      }
      else
         in=stdin;

      if (verbose)
      {
         fprintf(stderr,"\nTSAY Sound Converter For Sample Player\n");
         fprintf(stderr,"--------------------------------------\n\n");
         fprintf(stderr,"HEADER       : %s\n",name);
         fprintf(stderr,"SAMPLE RATE  : %0.3fkHz\n",SRATE(info.rate,channels)/1000);
         fprintf(stderr,"CHANNELS     : %d\n",channels);
         fprintf(stderr,"MAX FILE SIZE: %d\n\n",MAXSIZE);
      }

      MAXSIZE-=8;      /* File could pass limit by 8 bytes */

      if (channels==1)
         MonoConvert(in,&info,outfile);
      else
         StereoConvert(in,&info,outfile);

      if (infile[0]) fclose(in);
   }
   else
   {
      printf("\nTSAY SOUND CONVERTER FOR SAMPLE PLAYER\n");
      printf("          02/1996 Alvin Albrecht\n\n");
      printf("   This program converts 8 bit unsigned headerless\n");
      printf("sound samples into 4 bit packed samples with silence\n");
      printf("detection that can be played through the TS2068's\n");
      printf("AY-3-8912 sound chip using Sound Player.\n\n");
      printf("   You must specify the sample rate and whether the\n");
      printf("samples are mono or stereo.  This program is meant to\n");
      printf("be used in conjunction with SOX (available from\n");
      printf("ftp.cwi.nl/pub/audio/) which can convert sound samples\n");
      printf("in nearly all popular formats to the 8 bit unsigned\n");
      printf("samples that this program digests.\n\n");
      printf("USAGE: %s infile [options] outfile\n\n",argv[0]);
      printf("Where infile and outfile are filenames and can be\n");
      printf("replaced with - to indicate stdin or stdout respectively.\n\n");
      printf("OPTIONS:\n\n");
      printf("-n string    The text string is put into the header of\n");
      printf("             the converted sound record.  \"TSAY \" is\n");
      printf("             always present in the header.\n");
      printf("-r float     Indicates the sample rate in Hz.  Range is\n");
      printf("             MONO:  %0.3fkHz to %0.3fkHz\n",SRATE(256,1)/1000,SRATE(1,1)/1000);
      printf("             STRO:  %0.3fkHz to %0.3fkHz\n",SRATE(256,2)/1000,SRATE(1,2)/1000);
      printf("-c integer   # channels (1=mono, 2=stereo)\n");
      printf("-f integer   Maximum file size.  If the converted sound\n");
      printf("             record exceeds the number of bytes specified,\n");
      printf("             the record will be broken up into a maximum of\n");
      printf("             %d files.\n",MAXPARTS);
      printf("-V           Verbose mode.\n\n");
      printf("The defaults are -r 8000, -c 1, -f %d.\n\n",MAXSIZE);
   }
}
